@@ -19,13 +19,9 @@
 #include "olm/memory.hh"
 
 olm::Account::Account(
-) : next_one_time_key_id(0),
+) : num_fallback_keys(0),
+    next_one_time_key_id(0),
     last_error(OlmErrorCode::OLM_SUCCESS) {
-    // since we don't need to keep track of whether the fallback keys are
-    // published, use the published flag as in indication for whether the keys
-    // were generated
-    current_fallback_key.published = false;
-    prev_fallback_key.published = false;
 }
 
 
@@ -37,14 +33,14 @@ olm::OneTimeKey const * olm::Account::lookup_key(
             return &key;
         }
     }
-    if (current_fallback_key.published
+    if (num_fallback_keys >= 1
             && olm::array_equal(
                 current_fallback_key.key.public_key.public_key, public_key.public_key
             )
     ) {
         return &current_fallback_key;
     }
-    if (prev_fallback_key.published
+    if (num_fallback_keys >= 2
             && olm::array_equal(
                 prev_fallback_key.key.public_key.public_key, public_key.public_key
             )
@@ -67,14 +63,14 @@ std::size_t olm::Account::remove_key(
     }
     // check if the key is a fallback key, to avoid returning an error, but
     // don't actually remove it
-    if (current_fallback_key.published
+    if (num_fallback_keys >= 1
             && olm::array_equal(
                 current_fallback_key.key.public_key.public_key, public_key.public_key
             )
     ) {
         return current_fallback_key.id;
     }
-    if (prev_fallback_key.published
+    if (num_fallback_keys >= 2
             && olm::array_equal(
                 prev_fallback_key.key.public_key.public_key, public_key.public_key
             )
@@ -84,7 +80,7 @@ std::size_t olm::Account::remove_key(
     return std::size_t(-1);
 }
 
-std::size_t olm::Account::new_account_random_length() {
+std::size_t olm::Account::new_account_random_length() const {
     return ED25519_RANDOM_LENGTH + CURVE25519_RANDOM_LENGTH;
 }
 
@@ -120,7 +116,7 @@ static std::uint8_t * write_string(
 }
 
 
-std::size_t olm::Account::get_identity_json_length() {
+std::size_t olm::Account::get_identity_json_length() const {
     std::size_t length = 0;
     length += 1; /* { */
     length += sizeof(KEY_JSON_CURVE25519) - 1;
@@ -172,7 +168,7 @@ std::size_t olm::Account::get_identity_json(
 
 
 std::size_t olm::Account::signature_length(
-) {
+) const {
     return ED25519_SIGNATURE_LENGTH;
 }
 
@@ -193,7 +189,7 @@ std::size_t olm::Account::sign(
 
 
 std::size_t olm::Account::get_one_time_keys_json_length(
-) {
+) const {
     std::size_t length = 0;
     bool is_empty = true;
     for (auto const & key : one_time_keys) {
@@ -262,18 +258,19 @@ std::size_t olm::Account::mark_keys_as_published(
             count++;
         }
     }
+    current_fallback_key.published = true;
     return count;
 }
 
 
 std::size_t olm::Account::max_number_of_one_time_keys(
-) {
+) const {
     return olm::MAX_ONE_TIME_KEYS;
 }
 
 std::size_t olm::Account::generate_one_time_keys_random_length(
     std::size_t number_of_keys
-) {
+) const {
     return CURVE25519_RANDOM_LENGTH * number_of_keys;
 }
 
@@ -295,7 +292,7 @@ std::size_t olm::Account::generate_one_time_keys(
     return number_of_keys;
 }
 
-std::size_t olm::Account::generate_fallback_key_random_length() {
+std::size_t olm::Account::generate_fallback_key_random_length() const {
     return CURVE25519_RANDOM_LENGTH;
 }
 
@@ -306,19 +303,22 @@ std::size_t olm::Account::generate_fallback_key(
         last_error = OlmErrorCode::OLM_NOT_ENOUGH_RANDOM;
         return std::size_t(-1);
     }
+    if (num_fallback_keys < 2) {
+        num_fallback_keys++;
+    }
     prev_fallback_key = current_fallback_key;
     current_fallback_key.id = ++next_one_time_key_id;
-    current_fallback_key.published = true;
+    current_fallback_key.published = false;
     _olm_crypto_curve25519_generate_key(random, &current_fallback_key.key);
     return 1;
 }
 
 
 std::size_t olm::Account::get_fallback_key_json_length(
-) {
-    std::size_t length = 4 + sizeof(KEY_JSON_CURVE25519); /* {"curve25519":{}} */
-    OneTimeKey & key = current_fallback_key;
-    if (key.published) {
+) const {
+    std::size_t length = 4 + sizeof(KEY_JSON_CURVE25519) - 1; /* {"curve25519":{}} */
+    if (num_fallback_keys >= 1) {
+        const OneTimeKey & key = current_fallback_key;
         length += 1; /* " */
         length += olm::encode_base64_length(_olm_pickle_uint32_length(key.id));
         length += 3; /* ":" */
@@ -340,7 +340,7 @@ std::size_t olm::Account::get_fallback_key_json(
     pos = write_string(pos, KEY_JSON_CURVE25519);
     *(pos++) = '{';
     OneTimeKey & key = current_fallback_key;
-    if (key.published) {
+    if (num_fallback_keys >= 1) {
         *(pos++) = '\"';
         std::uint8_t key_id[_olm_pickle_uint32_length(key.id)];
         _olm_pickle_uint32(key_id, key.id);
@@ -356,6 +356,56 @@ std::size_t olm::Account::get_fallback_key_json(
     return pos - fallback_json;
 }
 
+std::size_t olm::Account::get_unpublished_fallback_key_json_length(
+) const {
+    std::size_t length = 4 + sizeof(KEY_JSON_CURVE25519) - 1; /* {"curve25519":{}} */
+    const OneTimeKey & key = current_fallback_key;
+    if (num_fallback_keys >= 1 && !key.published) {
+        length += 1; /* " */
+        length += olm::encode_base64_length(_olm_pickle_uint32_length(key.id));
+        length += 3; /* ":" */
+        length += olm::encode_base64_length(sizeof(key.key.public_key));
+        length += 1; /* " */
+    }
+    return length;
+}
+
+std::size_t olm::Account::get_unpublished_fallback_key_json(
+    std::uint8_t * fallback_json, std::size_t fallback_json_length
+) {
+    std::uint8_t * pos = fallback_json;
+    if (fallback_json_length < get_unpublished_fallback_key_json_length()) {
+        last_error = OlmErrorCode::OLM_OUTPUT_BUFFER_TOO_SMALL;
+        return std::size_t(-1);
+    }
+    *(pos++) = '{';
+    pos = write_string(pos, KEY_JSON_CURVE25519);
+    *(pos++) = '{';
+    OneTimeKey & key = current_fallback_key;
+    if (num_fallback_keys >= 1 && !key.published) {
+        *(pos++) = '\"';
+        std::uint8_t key_id[_olm_pickle_uint32_length(key.id)];
+        _olm_pickle_uint32(key_id, key.id);
+        pos = olm::encode_base64(key_id, sizeof(key_id), pos);
+        *(pos++) = '\"'; *(pos++) = ':'; *(pos++) = '\"';
+        pos = olm::encode_base64(
+            key.key.public_key.public_key, sizeof(key.key.public_key.public_key), pos
+        );
+        *(pos++) = '\"';
+    }
+    *(pos++) = '}';
+    *(pos++) = '}';
+    return pos - fallback_json;
+}
+
+void olm::Account::forget_old_fallback_key(
+) {
+    if (num_fallback_keys >= 2) {
+        num_fallback_keys = 1;
+        olm::unset(&prev_fallback_key, sizeof(prev_fallback_key));
+    }
+}
+
 namespace olm {
 
 static std::size_t pickle_length(
@@ -382,8 +432,8 @@ static std::uint8_t const * unpickle(
     std::uint8_t const * pos, std::uint8_t const * end,
     olm::IdentityKeys & value
 ) {
-    pos = _olm_unpickle_ed25519_key_pair(pos, end, &value.ed25519_key);
-    pos = olm::unpickle(pos, end, value.curve25519_key);
+    pos = _olm_unpickle_ed25519_key_pair(pos, end, &value.ed25519_key); UNPICKLE_OK(pos);
+    pos = olm::unpickle(pos, end, value.curve25519_key); UNPICKLE_OK(pos);
     return pos;
 }
 
@@ -414,9 +464,9 @@ static std::uint8_t const * unpickle(
     std::uint8_t const * pos, std::uint8_t const * end,
     olm::OneTimeKey & value
 ) {
-    pos = olm::unpickle(pos, end, value.id);
-    pos = olm::unpickle(pos, end, value.published);
-    pos = olm::unpickle(pos, end, value.key);
+    pos = olm::unpickle(pos, end, value.id); UNPICKLE_OK(pos);
+    pos = olm::unpickle(pos, end, value.published); UNPICKLE_OK(pos);
+    pos = olm::unpickle(pos, end, value.key); UNPICKLE_OK(pos);
     return pos;
 }
 
@@ -426,7 +476,8 @@ namespace {
 // pickle version 1 used only 32 bytes for the ed25519 private key.
 // Any keys thus used should be considered compromised.
 // pickle version 2 does not have fallback keys.
-static const std::uint32_t ACCOUNT_PICKLE_VERSION = 3;
+// pickle version 3 does not store whether the current fallback key is published.
+static const std::uint32_t ACCOUNT_PICKLE_VERSION = 4;
 }
 
 
@@ -437,8 +488,13 @@ std::size_t olm::pickle_length(
     length += olm::pickle_length(ACCOUNT_PICKLE_VERSION);
     length += olm::pickle_length(value.identity_keys);
     length += olm::pickle_length(value.one_time_keys);
-    length += olm::pickle_length(value.current_fallback_key);
-    length += olm::pickle_length(value.prev_fallback_key);
+    length += olm::pickle_length(value.num_fallback_keys);
+    if (value.num_fallback_keys >= 1) {
+        length += olm::pickle_length(value.current_fallback_key);
+        if (value.num_fallback_keys >= 2) {
+            length += olm::pickle_length(value.prev_fallback_key);
+        }
+    }
     length += olm::pickle_length(value.next_one_time_key_id);
     return length;
 }
@@ -451,8 +507,13 @@ std::uint8_t * olm::pickle(
     pos = olm::pickle(pos, ACCOUNT_PICKLE_VERSION);
     pos = olm::pickle(pos, value.identity_keys);
     pos = olm::pickle(pos, value.one_time_keys);
-    pos = olm::pickle(pos, value.current_fallback_key);
-    pos = olm::pickle(pos, value.prev_fallback_key);
+    pos = olm::pickle(pos, value.num_fallback_keys);
+    if (value.num_fallback_keys >= 1) {
+        pos = olm::pickle(pos, value.current_fallback_key);
+        if (value.num_fallback_keys >= 2) {
+            pos = olm::pickle(pos, value.prev_fallback_key);
+        }
+    }
     pos = olm::pickle(pos, value.next_one_time_key_id);
     return pos;
 }
@@ -463,28 +524,57 @@ std::uint8_t const * olm::unpickle(
     olm::Account & value
 ) {
     uint32_t pickle_version;
-    pos = olm::unpickle(pos, end, pickle_version);
+
+    pos = olm::unpickle(pos, end, pickle_version); UNPICKLE_OK(pos);
+
     switch (pickle_version) {
         case ACCOUNT_PICKLE_VERSION:
+        case 3:
         case 2:
             break;
         case 1:
             value.last_error = OlmErrorCode::OLM_BAD_LEGACY_ACCOUNT_PICKLE;
-            return end;
+            return nullptr;
         default:
             value.last_error = OlmErrorCode::OLM_UNKNOWN_PICKLE_VERSION;
-            return end;
+            return nullptr;
     }
-    pos = olm::unpickle(pos, end, value.identity_keys);
-    pos = olm::unpickle(pos, end, value.one_time_keys);
-    if (pickle_version == 2) {
+
+    pos = olm::unpickle(pos, end, value.identity_keys); UNPICKLE_OK(pos);
+    pos = olm::unpickle(pos, end, value.one_time_keys); UNPICKLE_OK(pos);
+
+    if (pickle_version <= 2) {
         // version 2 did not have fallback keys
-        value.current_fallback_key.published = false;
-        value.prev_fallback_key.published = false;
+        value.num_fallback_keys = 0;
+    } else if (pickle_version == 3) {
+        // version 3 used the published flag to indicate how many fallback keys
+        // were present (we'll have to assume that the keys were published)
+        pos = olm::unpickle(pos, end, value.current_fallback_key); UNPICKLE_OK(pos);
+        pos = olm::unpickle(pos, end, value.prev_fallback_key); UNPICKLE_OK(pos);
+        if (value.current_fallback_key.published) {
+            if (value.prev_fallback_key.published) {
+                value.num_fallback_keys = 2;
+            } else {
+                value.num_fallback_keys = 1;
+            }
+        } else  {
+            value.num_fallback_keys = 0;
+        }
     } else {
-        pos = olm::unpickle(pos, end, value.current_fallback_key);
-        pos = olm::unpickle(pos, end, value.prev_fallback_key);
+        pos = olm::unpickle(pos, end, value.num_fallback_keys); UNPICKLE_OK(pos);
+        if (value.num_fallback_keys >= 1) {
+            pos = olm::unpickle(pos, end, value.current_fallback_key); UNPICKLE_OK(pos);
+            if (value.num_fallback_keys >= 2) {
+                pos = olm::unpickle(pos, end, value.prev_fallback_key); UNPICKLE_OK(pos);
+                if (value.num_fallback_keys >= 3) {
+                    value.last_error = OlmErrorCode::OLM_CORRUPTED_PICKLE;
+                    return nullptr;
+                }
+            }
+        }
     }
-    pos = olm::unpickle(pos, end, value.next_one_time_key_id);
+
+    pos = olm::unpickle(pos, end, value.next_one_time_key_id); UNPICKLE_OK(pos);
+
     return pos;
 }
